/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

/* Uses games package to compute values of corridors in go.  A
   position represents a (white) tree of invading groups, each of
   which needs one move to connect to its parent, whose other
   liberties are (black) "blocked" and "unblocked" corridors of
   territory. -- Best expained by example. */

/*
  ....|.|
  ...XXXO...
  ...X.X.X|.
  |..X.X.XX.
  OOOX.XOOXX
  O..OOO.O.X
*/

/* Here every group with a | next to it is alive (immortal).  There
   are two white invading groups:  The first is 3 stones attacking a
   "blocked" corridor of length 3 and needs one move to connect to
   life.  The second is 3 stones attacking an "unblocked" corridor of
   length 2 and a "blocked" corridor of length 1, and needs one move
   to connect to the first invading group. */

/* The input format for this position is
   corridor (3b3 (3b1u2))
   In general, there is a left paren, followed by the size of a
   group followed by a list of liberties, followed by a close paren.
   Each liberty can either be a blocked corridor -- a "b" followed by
   a length, or an unblocked (u), or another invading group which
   needs to connect to the current ((3b1u2) in the current example).
   */

extern int corridor_get(list_type);
extern game_type corridor_val(int);
extern void corridor_printf (int);
extern void corridor_clear (void);

#define F1		(1<<(INT_SIZE-1))
#define F0		(1<<(INT_SIZE-2))
#define FLAGS		(F0 | F1)
#define SIZE		(0)
#define BLOCKED		(F0)
#define UNBLOCKED	(F1 | F0)
#define FOLLOWER	(F1)
#define DATA		(F0 - 1)
#define MAX_DEPTH 10
#define MAX_INVASIONS (MAX_SIZE < DATA ? MAX_SIZE : DATA)
