/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

#include <stdlib.h>
#include <stdio.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "operations.h"
#include "output.h"
#include "bitmatrix.h"
#include "domino.h"

/* To the 'student' eager to learn how to implement your own game:
 * First review function domino() below.  It calls input_domino, and
 * domino_position().  Naturally, input_domino() gets the input.
 * domino_position() evaluates the position by placing domino's, and
 * breaking the remaining position into its connected components.
 * A hash table is used to store previously computed results as a
 * means of dynamic programming, to redundant recursive calls.
 */

game_type place_domino (B_type, int, int, int, int);
void copy_connected_component_at (B_type, B_type, int, int);

game_type domino_position (B_type pos)
{
	game_type g;
	int i, j;
	list_type left, right, l;
	
	l = B_to_list (pos);
	if (hash_test (DOMINO_KEY, l)) return list_free (l), hash_get_last();
	left=list_make();
	right=list_make();
	for (i = 0; i < B_x(pos); i++)
		for (j = 0; j < B_y(pos); j++)
			if (B_get (pos, i, j)) {
				if ((i+1 < B_x(pos)) && B_get(pos, i+1, j))
					list_insert (right, place_domino (pos, i,j, i+1,j));
				if ((j+1 < B_y(pos)) && B_get(pos, i, j+1))
					list_insert (left, place_domino (pos, i,j, i,j+1));
			}
	g = make (left, right);
	hash_put (DOMINO_KEY, l, g);
	return g;
}

game_type place_domino (B_type pos, int x1, int y1, int x2, int y2)
{
	game_type g;
	B_type new, old;
	int i, j;
	boolean done_something;

	g = zero;
	old = B_copy (pos);
	B_clear (old, x1, y1);
	B_clear (old, x2, y2);
	for (done_something=TRUE;done_something;) {
		done_something=FALSE;
		for (j = 0; j < B_y(old); j++)
			for (i = 0; i < B_x(old); i++)
				if (B_get (old, i, j)) {
					new = B_make(B_x(old),B_y(old));
					copy_connected_component_at (old, new, i, j);
					new = B_pack (new);
					g = plus (g, domino_position (new));
					B_free (new);
					done_something = TRUE;
				}
	}
	B_free (old);
	return g;
}

int pos_x[]={1,0,-1,0};
int pos_y[]={0,1,0,-1};
void copy_connected_component_at (B_type old, B_type new, int x, int y)
{
	int i,j,n;

	B_clear (old, x, y);
	B_set (new, x, y);
	for (n = 0; n <= 3; n++) {
		i = x + pos_x[n];
		j = y + pos_y[n];
		if (i >= 0 && j >= 0 && i < B_x(old) && j < B_y(old) && B_get (old,i,j))
			copy_connected_component_at (old, new, i, j);
	}
}

B_type domino_input(void)
{
	B_type pos;
	char s[MAXBUFF];
	int i=0, j;
	
	pos = B_make(DOMINO_INPUT_SIZE,DOMINO_INPUT_SIZE);
	printf ("Enter domineering position followed by extra <cr>\n");
	for (j=0; gets(s) != NULL && *s != '\0';j++) {
		for (i=0; *(s+i) != '\0'; i++)
			if (*(s+i) != ' ') B_set (pos, i, j);
	}
	if (j==0 && i==0) return (B_type) NULL;
	else return B_pack(pos);
}

void domino_print (B_type pos) {
	int i, j;
	
	for (j = 0; j < B_y(pos); j++) {
		for (i = 0; i < B_x(pos); i++)
			printf (B_get (pos, i, j) ? "x" : " ");
		printf ("\n");
	}
}
	
/* FUNCTION TO AID IN UNDOCUMENTED BEHAVIOR  (If you want to learn about how
   to implement your own game, ignore it) */
static int dc_x (int x, int y, int i) {
	int m, n;
	if (x == 0) { /* domineering chain hardwired: over : 3x4x3x3 */  
		n = i%9;
		return n <= 2 ? n :
		       n <= 5 ? 2 :
		       n <= 7 ? 7-n :
			        0;
	}
	m = i/(x+y-2);
	n = i%(x+y-2);
	if (n<x-1) return (x-1)*m + n;
	else return (x-1)*(m+1);
}

/* ALSO UNDOCUMENTED */
static int dc_y (int x, int y, int i) {
	int m, n;
	if (x == 0) {
		m = i/9;
		n = i%9;
		return n <= 2 ? m*5 :
		       n <= 5 ? m*5 + n-2 :
		       n <= 7 ? m*5 + 3 :
			        m*5 + n-4;
	}
	m = i/(x+y-2);
	n = i%(x+y-2);
	if (n<x-1) return (y-1)*m;
	else return (y-1)*m + (n-(x-1));
}

#define X(i) dc_x (x,y,(i))
#define Y(i) dc_y (x,y,(i))
/* ALSO UNDOCUMENTED */
game_type domino_chain (int x, int y, int from, int to) {
	B_type a;
	int i, fx, fy;
	game_type g;

	if (to <= from) return zero;
	fx = (x==0 ? 0 : X(from));
	fy = Y(from);
	a = B_make (x == 0 ? 3 : X(to)-fx+1, Y(to)-fy+1);
	for (i = from; i <= to; i++)
		B_set (a, X(i)-fx, Y(i)-fy);
	g = domino_position (a);
	B_free (a);
	return g;
}
	
game_type domino (list_type l)
{
	int x, y, from, to, length, last, width;
	B_type pos;
	char s[MAXBUFF];
	game_type g;
	
	length = list_length (l);
/* HERE'S WHERE THE ACTION IS */
	if (length == 0) {
		pos = domino_input();
		g = domino_position (pos);
		B_free (pos);
		return g;
	}
/* THE REMAINDER OF THIS FUNCTION IS UNDOCUMENTED BEHAVIOR */
	if (length == 1) {
		myerror ("Domineering position with 1 arg");
		return NO_GAME;
	}
	last=0;
	x = list_nth (l, 1);
	y = list_nth (l, 2);
	if (length == 2) {
		width = (x == 0 ? 9 : x+y-2);
		printf ("\n%dx%d domineering chain - period %d, saltus %s", x, y, period, game_sprintn (s,saltus, 20));
		if (flags & EVALUATE_FLAG) printf (", warming function: %s", warm_string);

		else printf ("\n");
		printf ("%3s |",""); for (from=0; from < width; from++) printf (" %8d", from+1); printf ("\n");
		printf ("---------------------------------------------------------------------------------------\n");
		for (to=0; (to <= last*3 || to <= 10) && to <= 200; to++) {
			printf ("%3d |", to+1);
			for (from=0; from < width; from++) {
				g = domino_chain (x,y,from,to);
				if (to>period && eq (g, plus (domino_chain (x,y,from,to-period), saltus)))
					*s='\0';
				else {
					game_sprintn (s, g, MAXBUFF);
					last = to;
				} printf (" %8s", s);
			}
			printf ("\n");
			fflush (stdout);
		}
		return NO_GAME;
	}
	from = list_nth (l, 3);
	if (length == 3) {
		printf ("%3s |  %d by %d domineering chain starting with the %d'th square\n", " ", x, y, from);
		printf ("---------------------------------------------------------------------------------------\n");
		from--;
		for (to=0; (to<=last*3 || to<=10) && to < 100; to++) {
			g = domino_chain (x,y,from,to);
			if (to>period && eq (g, plus (domino_chain (x,y,from,to-period), saltus))) *s = '\0';
			else game_sprintn (s, g, MAXBUFF), last=to;
			printf ("%3d | %s", to+1, s);
			printf ("\n");
			fflush (stdout);
		}
		return NO_GAME;
	}
	to = list_nth (l, 4);
	if (length == 4)
		return domino_chain (x,y,from,to);
	myerror ("domino with more than 4 arguments");
	return NO_GAME;
}
