/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

#ifndef IBMPC
/* Macros PRINT_COMMAND and HELP_FILE used to be here.  Now they are */
 /* specified in the Makefile.  You no longer need to specify them. */
 /* On PC's (if the program works at all), you should not set up these */
 /* macros --- help just doesn't work on non-UNIX beasts. */
#endif

#include <string.h>


#ifdef Ibm
#define MAX_SIZE 3000
#define INT_SIZE 16
#else
#define MAX_SIZE 32000
#define INT_SIZE 32
#endif

#define FALSE 0
#define TRUE 1

#define MAXBUFF 1000

/* print options */
#define INTEGERS_FLAG (1<<0)
#define NUMBERS_FLAG (1<<1)
#define STARS_FLAG (1<<2)
#define UPS_FLAG (1<<3)
#define TINYS_FLAG (1<<4)
#define ZEROS_FLAG (1<<16)
#define ONS_FLAG (1<<12)
#define VARS_FLAG (1<<5)
#define EXPAND_FLAG (1<<6)
#define PROMPT_FLAG (1<<7)
/* check if g = h.1* */
#define EVALUATE_FLAG (1<<8)
#define NORMALIZE_FLAG (1<<14)
#define ECHO_INPUT_FLAG (1<<9)
/* make unproven assumptions (corridors.c) for efficiency */
#define ASSUME_FLAG (1<<10)
/* make false assumptions (corridors.c) for efficiency -- in particular that the size of
   an invading group is irrelevent */
#define ASSUME2_FLAG (1<<15)
/* reinitialize hash table between commands */
#define CLEAN_FLAG (1<<11)
/* output formulas in TeX */
#define TEX_FLAG (1<<13)
#define TRACE_FLAG (1<<17)	/* trace evaluation */
#define ISH_FLAG (1<<18)	/* print "-ish" values */
#define ALL_FLAG (INTEGERS_FLAG|NUMBERS_FLAG|STARS_FLAG|UPS_FLAG|TINYS_FLAG|ONS_FLAG|\
                  VARS_FLAG|EXPAND_FLAG|PROMPT_FLAG|NORMALIZE_FLAG|ZEROS_FLAG)

/* hash table entries */
#define UNEXPENDABLE_KEYS	(1 << 15)
#define EXPENDABLE_KEYS		(1 << 14)
#define OTHER_KEYS		(1 << 13)
#define SIMPLIFY_KEY		( 1 | UNEXPENDABLE_KEYS)
#define PLUS_KEY		( 2 | EXPENDABLE_KEYS)
#define NEGATIVE_KEY		( 3 | EXPENDABLE_KEYS)
#define GE_KEY			( 4 | EXPENDABLE_KEYS)
#define NUM_UP_STAR_BABYKO_KEY	( 5 | EXPENDABLE_KEYS)
#define COOL_KEY		( 8 | EXPENDABLE_KEYS)
#define TEMPERATURE_KEY		( 9 | EXPENDABLE_KEYS)
#define FREEZE_KEY		(10 | EXPENDABLE_KEYS)
#define HEAT_KEY		(12 | EXPENDABLE_KEYS)
#define MULT_KEY		(13 | EXPENDABLE_KEYS)
#define AW_KEY			(14 | EXPENDABLE_KEYS)
#define VAR_KEY			(15 | UNEXPENDABLE_KEYS)
#define CORRIDOR_KEY		(16 | OTHER_KEYS)
#define CORRIDOR_VALUE_KEY	(18 | EXPENDABLE_KEYS)
#define INT_KO_STAR_KEY		(19 | UNEXPENDABLE_KEYS)
#define INT_BABYKO_KEY		(20 | UNEXPENDABLE_KEYS)
#define DOMINO_KEY		(21 | EXPENDABLE_KEYS)
#define KONANE_KEY		(22 | EXPENDABLE_KEYS)
#define TOADS_KEY		(23 | EXPENDABLE_KEYS)
#define DOMSTR_KEY		(24 | EXPENDABLE_KEYS)
#define SOWING_KEY		(25 | EXPENDABLE_KEYS)
#define ISOWING_KEY		(26 | EXPENDABLE_KEYS)
#define MANCALA_KEY		(27 | EXPENDABLE_KEYS)
#define REVMANCALA_KEY		(28 | EXPENDABLE_KEYS)
#define HOTMANCALA_KEY		(29 | EXPENDABLE_KEYS)
#define COLDMANCALA_KEY		(30 | EXPENDABLE_KEYS)
#define WARI_KEY		(31 | EXPENDABLE_KEYS)
#define IWARI_KEY		(32 | EXPENDABLE_KEYS)

/* These indicate what to do when hash table needs space.
   OTHER_KEYS are sometimes expendable, depending on context */

/* ARGUMENT_SYMBOL is the internal string representation of the argument to 
   cool_function and warm_function. */
#define ARGUMENT_SYMBOL "\1"

typedef int boolean;
typedef unsigned game_type;
typedef game_type (* function_type)();
typedef boolean (* test_type) (game_type, game_type);

extern void		help();
extern void		stat();
extern void		myerror (char *);
extern int		game_parse(char *);
extern char **		envp;
extern unsigned long	flags;

extern char help_file[];
extern void _exit(int), perror(const char *), execle();
extern int printf(const char *, ...), fork(), wait(), yyparse(), fflush();
extern char *gets();
#ifdef __TURBOC__
extern void exit(int), free(void *);
extern long coreleft(void);
#endif

extern int tlevel;	/* Trace level */
