/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "externs.h"
#include "rational.h"
#include "list.h"
#include "hash.h"
#include "operations.h"
#include "output.h"

char			help_file[MAXBUFF];
unsigned long		flags = ALL_FLAG;
char 			**envp;
int tlevel = 0;

void myerror(char *s) {
	perror (s);
}

void help() {
#ifdef HELP_FILE
	char help_filename[MAXBUFF];
	FILE *infile;

	strcpy (help_filename, HELP_FILE);
	if (strlen (help_file) > 0) {
		strcat (help_filename, "_");
		strcat (help_filename, help_file);
	}
	if ((infile = fopen (help_filename, "r")) == NULL)
		if (strlen (help_file) > 1)
			printf ("No help available on that topic.\n");
		else
			printf ("Help seems to be broken on this system.  Sorry.\n");
	else {
#ifdef PRINT_COMMAND		
		fclose (infile);
		switch (fork ()) {
		case -1:
			perror ("Sorry, couldn't fork");
			break;
		case 0:
			execle (PRINT_COMMAND, PRINT_COMMAND, help_filename, 0, envp);
			perror ("Sorry, couldn't exec");
			_exit (-1);
		default:
			if (wait(0) < 0)
				perror ("wait in help");
		}
#else
		{
			char line[MAXBUFF];
			while (fgets (line, MAXBUFF, infile) != '\0')
				printf ("%s", line);
			fclose (infile);
		}
#endif
	}
#else
	printf ("Help seems to be broken on this system.  Sorry\n");
#endif
}
	



extern char *yysptr;
extern char yysbuf[];
int game_parse (char *s) {
	parser_input = s;
	parser_output = NO_GAME;
	if (flags & CLEAN_FLAG) clean();
	yysptr = yysbuf;	/* Initialize lex's unput buffer to empty in case 
				 * last call to yyparse() bombed. */
	if (yyparse()) return NO_GAME;
	else return parser_output;
}

extern unsigned long game_stat();
extern unsigned long hash_stat();
extern unsigned long list_stat();
extern unsigned long corridor_stat();

void stat (void) {
#ifdef __TURBOC__
	printf ("total lists: %lu; coreleft %lu\n",
		hash_stat() + game_stat() + corridor_stat() + list_stat(), coreleft());
#else
	printf ("total lists: %lu\n",
		hash_stat() + game_stat() + corridor_stat() + list_stat());
#endif
}

void games (void) {
	char input[MAXBUFF];
	game_type g;

	printf ("Type help, if you need it, control-D to exit\n");
	for (;;) {
		if (flags & PROMPT_FLAG) printf ("> ");
		if (gets (input) == NULL || !strncmp (input,"quit",4) || !strncmp (input,"exit",4)) {
			printf ("\n");
			exit(0);	/* eof */
		}
		if (flags & ECHO_INPUT_FLAG) printf ("%60s => ", input);
		fflush (stdout);
		strcat (input, "\n");
		g = game_parse (input);
		if (g) set_var (dummy_var = make_var ("$"), g);
		if (g) game_printf (g);
		fflush (stdout);
	}
}

