/* Copyright (C) David Wolfe, 1990.  All rights reserved. */
/* Includes mostly code written by David Moews */

%{
#include "externs.h"
#include "rational.h"
#include "list.h"
#include "operations.h"
#include "output.h"
#include "corridors.h"
#include "bitmatrix.h"
#include "domino.h"
#include "konane.h"
#include "toads.h"
#include "domstr.h"
#include "sowing.h"
#include "mancala.h"
#include "wari.h"

#define vsl(a,b) make(list_copy_1((a)), (b))
#define vsr(a,b) make((a), list_copy_1((b)))

boolean non_syn_error = FALSE;
%}

%start line

%union {
	int ival;
	Q_type qval;
	list_type lval;
	game_type gval;
        var_type vval;
}

%type <vval> vs
%type <ival> mask c ce
%type <lval> l cl nl
%type <gval> num atom atoms numatoms term
%type <gval> g gt gts line gop gterm
%type <gval> gp		/* plain g (w/o bars in outer level) */
%type <gval> gb gb1 gb2 gb3 gb4

%token <ival> NUM ZEROS BARZEROS ONNUM
%token <vval> S
%token EOS
%token FREEZE CHILL MEAN TEMPERATURE AW ON OFF TWEAK
%token HELP RESET STAT
%token INTEGERS NUMBERS STARS UPS TINYS ONS VARS EXPAND ALL EVALUATE NORMALIZE PROMPT ECHO_ ASSUME ASSUME2 CLEAN TEX ZEROS
%token BAR BARR BARRR BARRRR BARRRRR NO
%token MINUS_K MINUS_k
%token TINY MINY
%token UNKNOWN
%token CORRIDOR DOMINO KONANE TOADS DOMSTR SOWING ISOWING
%token MANCALA REVMANCALA HOTMANCALA COLDMANCALA WARI IWARI
%token SALTUS PERIOD TRACE ISH

%%
line	: EOS /*empty*/	{ }
	| error EOS	{if (!non_syn_error)
    				    myerror("Syntax error.");
                                 non_syn_error = FALSE;
				 yyerrok;
   				}
	| g EOS		{parser_output = $1; YYACCEPT;}
	| g '?' g EOS	{(void)printf("%s\n", compare($1, $3));
					parser_output = NO_GAME;}
	| TRACE NUM EOS	{flags |= TRACE_FLAG; tlevel = $2;}
	| mask EOS	{flags |= $1;}
	| NO mask EOS 	{flags &= ~$2;}
	| HELP EOS	{(void) help();}
	| STAT EOS	{stat();}
	| RESET EOS	{reinit();}
	| DOMINO nl EOS		{parser_output = domino ($2);}
	| KONANE nl EOS		{parser_output = konane ($2);}
        | TOADS EOS		{parser_output = toads();}
        | DOMSTR nl EOS		{parser_output = domstr($2);}
        | SOWING EOS		{parser_output = sowing(0,0,0);}
        | ISOWING EOS		{parser_output = sowing(1,0,0);}
        | SOWING NUM NUM EOS	{parser_output = sowing(0,$2,$3);}
        | ISOWING NUM NUM EOS	{parser_output = sowing(1,$2,$3);}
        | MANCALA EOS		{parser_output = mancala(0,0);}
        | REVMANCALA EOS	{parser_output = revmancala(0,0);}
        | HOTMANCALA EOS	{parser_output = hotmancala(0,0);}
        | COLDMANCALA EOS	{parser_output = coldmancala(0,0);}
        | MANCALA NUM NUM EOS		{parser_output = mancala($2,$3);}
        | REVMANCALA NUM NUM EOS	{parser_output = revmancala($2,$3);}
        | HOTMANCALA NUM NUM EOS	{parser_output = hotmancala($2,$3);}
        | COLDMANCALA NUM NUM EOS	{parser_output = coldmancala($2,$3);}
        | WARI nl EOS		{parser_output = wari(0,$2);}
        | IWARI nl EOS		{parser_output = wari(1,$2);}
	| SALTUS EOS		{parser_output = saltus; YYACCEPT;}
	| SALTUS g EOS		{saltus = $2;}
	| PERIOD EOS		{parser_output = g_int (period); YYACCEPT;}
	| PERIOD NUM EOS	{period = $2;}
	;	

mask	: INTEGERS		{$$ = INTEGERS_FLAG;}
	| NUMBERS		{$$ = NUMBERS_FLAG;}
	| STARS			{$$ = STARS_FLAG;}
	| UPS			{$$ = UPS_FLAG;}
	| TINYS			{$$ = TINYS_FLAG;}
	| ZEROS			{$$ = ZEROS_FLAG;}
	| ONS			{$$ = ONS_FLAG;}
	| EXPAND		{$$ = EXPAND_FLAG;}
	| VARS			{$$ = VARS_FLAG;}
	| ALL			{$$ = ALL_FLAG;}
	| PROMPT		{$$ = PROMPT_FLAG;}
	| EVALUATE		{$$ = EVALUATE_FLAG;}
	| NORMALIZE		{$$ = NORMALIZE_FLAG;}
	| ECHO_			{$$ = ECHO_INPUT_FLAG;}
	| ASSUME		{$$ = ASSUME_FLAG;}
	| ASSUME2		{$$ = ASSUME2_FLAG;}
	| CLEAN			{$$ = CLEAN_FLAG;}
	| TEX			{$$ = TEX_FLAG;}
	| TRACE			{$$ = TRACE_FLAG;}
	| ISH			{$$ = ISH_FLAG;}
	;

nl	: 			{$$ = list_make();}
	| NUM nl		{list_insert_unsorted ($2, $1), $$ = $2;}
	 

c	: '(' cl ')'				{$$ = corridor_get ($2);
						list_free ($2);}
	;

cl	: ce					{$$ = list_copy_1 ($1);}
	| cl ce					{list_insert ($1, $2); $$ = $1;}
	;

ce	: NUM					{$$ = SIZE | $1;}
	| 'b' NUM				{$$ = BLOCKED | $2;}
	| 'u' NUM				{$$ = UNBLOCKED | $2;}
	| c					{$$ = FOLLOWER | $1;}
	;

l 	: gt					{$$ = list_copy_1($1);}
	| l ',' gt				{list_insert ($$ = $1, $3);}
	;

gb1	: l BAR l				{$$ = make($1, $3);}
	| l BAR gb1				{$$ = vsr($1, $3);}
	| ZEROS BAR gt				{$$ = zeros_vs ($3, $1);}
	| gt BARZEROS				{$$ = vs_zeros ($1, $2);}
	;

gb2	: gb1					{$$ = $1;}
	| l BARR l				{$$ = make($1, $3);}
	| l BARR gb2				{$$ = vsr($1, $3);}
	| gb1 BARR l				{$$ = vsl($1, $3);}
	| gb1 BARR gb2				{$$ = vs ($1, $3);}
	;

gb3	: gb2					{$$ = $1;}
	| l BARRR l				{$$ = make ($1, $3);}
	| l BARRR gb3				{$$ = vsr($1, $3);}
	| gb2 BARRR l				{$$ = vsl($1, $3);}
	| gb2 BARRR gb3				{$$ = vs ($1, $3);}
	;

gb4	: gb3					{$$ = $1;}
	| l BARRRR l				{$$ = make ($1, $3);}
	| l BARRRR gb4				{$$ = vsr($1, $3);}
	| gb3 BARRRR l				{$$ = vsl($1, $3);}
	| gb3 BARRRR gb4			{$$ = vs ($1, $3);}
	;

gb 	: gb4					{$$ = $1;}
	| l BARRRRR l				{$$ = make ($1, $3);}
	| l BARRRRR gb				{$$ = vsr($1, $3);}
	| gb4 BARRRRR l				{$$ = vsl($1, $3);}
	| gb4 BARRRRR gb			{$$ = vs ($1, $3);}
	;

g	: vs '=' g				{set_var($1, $$ = $3);}
	| gts;
	;

gts	: gterm					{$$ = $1;}
	| gts '+' gterm				{$$ = plus($1, $3);}
	| gts '-' gterm				{$$ = minus($1, $3);}
	;

gterm	: gt					{$$ = $1;}
	| gb					{$$ = $1;}
	;

gt	: gop					{$$ = $1;}
	| gp					{$$ = $1;}
	| gp '[' g ']'				{$$ = cool ($1, mean ($3));}
	| gp '<' g '>'				{$$ = gexp ($1, Q_p (mean ($3)));}
	| gp '.' gp				{$$ = mult($1, $3);}
	| gp ONNUM				{$$ = on ($1, $2);}
	| gt gop				{$$ = plus ($1,$2);}
	;

gop	: FREEZE gp				{$$ = freeze ($2);}
	| TWEAK gp				{$$ = tweak($2);}
	| '%' gp				{$$ = overheat ($2, NO_GAME, NO_GAME);}
	| '%' '<' g '>' gp			{$$ = heat ($5, $3);}
	| '%' '[' g ']' '<' g '>' gp		{$$ = overheat ($8, $3, $6);}
	| ON '[' g ']' '<' g '>'		{$$ = on ($3, Q_p (mean ($6)));}
	| OFF '[' g ']' '<' g '>'		{$$ = off ($3, Q_p (mean ($6)));}
	| CORRIDOR c				{$$ = corridor_val ($2);}
	| MEAN gp				{$$ = num (mean ($2));}
	| TEMPERATURE gp			{$$ = num (temperature ($2));}
	| AW gp					{$$ = atomic_weight ($2);}
	;

num	: NUM					{$$ = num(Q_int($1));}
	| NUM '/' NUM				{$$ = num(Q($1, $3));}
	| '-' NUM '/' NUM			{$$ = num(Q(-$2, $4));}
	| NUM NUM '/' NUM			{$$ = num(Q_plus(Q_int ($1), Q($2, $4)));}
	| '-' NUM NUM '/' NUM			{$$ = num(Q_plus(Q_int (-$2), Q(-$3, $5)));}
	| '-' vs				{$$ = negative (get_var ($2));}
	| '-' NUM				{$$ = num(Q_int(-$2));}
	;

atom	: '*'					{$$ = star(1);}
	| '*' NUM				{$$ = star($2);}
	| '^'					{$$ = up(1);}
	| '^' NUM				{$$ = up($2);}
	| 'v'					{$$ = down(1);}
	| 'v' NUM				{$$ = down($2);}
	| TINY '[' g ']'			{$$ = tiny ($3);}
	| MINY '[' g ']'			{$$ = miny ($3);}
	| 'K'					{$$ = ko;}
	| MINUS_K				{$$ = kobar;}
	| 'k'					{$$ = babyko;}
	| MINUS_k				{$$ = babykobar;}
	| '{' g '}'				{$$ = $2;}
	| '(' g ')'				{$$ = $2;}
	;

atoms	: atom					{$$ = $1;}
	| atoms atom				{$$ = plus ($1, $2);}
	;

numatoms: atoms					{$$ = $1;}
	| num					{$$ = $1;}
	| num atoms				{$$ = plus($1, $2);}
	;

term	: numatoms				{$$ = $1;}
/*	| numatoms '%' numatoms			{$$ = plus ($1, mult ($3, one_star));}*/
	;

gp	: term					{$$ = $1;}
	| vs					{$$ = get_var($1);
						 list_free ($1);
						 if ($$ == NO_GAME) {
                                                   myerror("Variable not set.");
                                                   non_syn_error = TRUE;
                                                   YYERROR;
                                                 }
						}
	| '{' l BAR '}'				{$$ = make($2, list_make());}
	| '{' BAR l '}'				{$$ = make(list_make(), $3);}
	| '{' BAR '}'				{$$ = zero;}
	| '{' l BARR '}'			{$$ = make($2, list_make());}
	| '{' gb1 BARR '}'			{$$ = vsl($2, list_make());}
	| '{' BARR l '}'			{$$ = make(list_make(), $3);}
	| '{' BARR gb1 '}'			{$$ = vsr(list_make(), $3);}
	| '{' BARR '}'				{$$ = zero;}
	| '{' l BARRR '}'			{$$ = make($2, list_make());}
	| '{' gb2 BARRR '}'			{$$ = vsl($2, list_make());}
	| '{' BARRR l '}'			{$$ = make(list_make(), $3);}
	| '{' BARRR gb2 '}'			{$$ = vsr(list_make(), $3);}
	| '{' BARRR '}'				{$$ = zero;}
	| '{' l BARRRR '}'			{$$ = make($2, list_make());}
	| '{' gb3 BARRRR '}'			{$$ = vsl($2, list_make());}
	| '{' BARRRR l '}'			{$$ = make(list_make(), $3);}
	| '{' BARRRR gb3 '}'			{$$ = vsr(list_make(), $3);}
	| '{' BARRRR '}'			{$$ = zero;}
	| '{' l BARRRRR '}'			{$$ = make($2, list_make());}
	| '{' gb4 BARRRRR '}'			{$$ = vsl($2, list_make());}
	| '{' BARRRRR l '}'			{$$ = make(list_make(), $3);}
	| '{' BARRRRR gb4 '}'			{$$ = vsr(list_make(), $3);}
	| '{' BARRRRR '}'			{$$ = zero;}
	;

vs	: S					{$$ = $1;}
	;
%%
char *parser_input;
game_type parser_output;

void yyerror (s)
	char *s;
{
}

#include "parse.c"
