/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

/* Kluges for kos (not well thought out -- certainly unproven)
 * are marked with the comment *KLUGED KO* */

#include <stdlib.h>
#include "externs.h"
#include "rational.h"
#include "list.h"
#include "hash.h"
#include "operations.h"
#include "output.h"
#include "corridors.h"

#define GNIL ((game_cell_type) 0)
#define nusb(n,u,s,k) num_up_star_babyko(n,u,s,k)
#define is_nusb(g) is_num_up_star_babyko(g)

game_type		saltus;
int			period;
game_cell_type		game_cell[MAX_GAMES];	/* array of pointers to game cells */
static unsigned long	game_cells = 0;		/* Total number of game cells used */
static game_cell_type	game_free_list = GNIL;
static Q_type		frozen_temperature;
static game_type	unit;
static list_type	unit_adder = NIL, unit_neg_adder = NIL;
static int		_rstar_;

game_type zero;
game_type one;
game_type two;
game_type neg1;
game_type one_star;
game_type ko;
game_type kobar;
game_type babyko;
game_type babykobar;
var_type  dummy_var;	/* set_var shouldn't make game point to variable */

static void game_free_cell (game_cell_type cell) {
	cell -> next = game_free_list;
	game_free_list = cell;
	return;
}

static void game_free (game_type g) {
	list_free (game_cell[g]->left);
	list_free (game_cell[g]->right);
	game_free_cell (game_cell[g]);
	if (g == game_cells) game_cells--;
	return;
}

static game_cell_type game_get_cell (void) {
	game_cell_type cell, p;
	unsigned i;
	
	if (game_free_list == GNIL) {
		if (!(game_free_list = (game_cell_type)
		      malloc (GAME_BLOCK_SIZE * sizeof (struct game_struct))))
			myerror ("game_get_cell: malloc couldn't free block");
		for (i = 1, p=game_free_list; i < GAME_BLOCK_SIZE; i++, p++)
			p->next = p+1;
		p->next = GNIL;
	}
	cell = game_free_list;
	game_free_list = game_free_list->next;
	return cell;
}

static game_type new_game ()
{
	if (++game_cells >= MAX_GAMES) {
		myerror ("new_game: No more games.  Sorry.");
		exit (1);
	}
	game_cell[game_cells] = game_get_cell();
	game_cell[game_cells]->left = NIL;
	game_cell[game_cells]->right = NIL;
	game_cell[game_cells]->flags = 0;
	game_cell[game_cells]->Num = Q_0;
	game_cell[game_cells]->Star = 0;
	game_cell[game_cells]->Up = 0;
	game_cell[game_cells]->var = NIL;
	game_cell[game_cells]->next = GNIL;
	return (game_type) game_cells;
}

static game_type dominates (list_type list, game_type g, test_type test) {
	list_type l;
	
	for (l = list_start (list); l != NIL; l = list_rest (l))
		if ((* test) (list_pick (l), g)) return list_pick (l);
	return (game_type) FALSE;
}


static void reverse_reversals (game_type g) {
	game_type h;
	list_type l;
	list_type l2;
	
	for (l = list_start (left_options (g)); l != NIL;) {
		h = dominates (right_options (list_pick (l)), g, le);
		if (h) {
			list_delete (left_options (g), list_pick (l));
			for (l2 = list_start (extended_left_options (h)); l2 != NIL; l2 = list_rest (l2))
				list_insert (left_options (g), list_pick(l2));
			l = list_start (left_options (g));
		}
		else
			l = list_rest (l);
	}
	for (l = list_start (right_options (g)); l != NIL;) {
		h = dominates (left_options (list_pick (l)), g, ge);
		if (h) {
			list_delete (right_options (g), list_pick (l));
			for (l2 = list_start (extended_right_options (h)); l2 != NIL; l2 = list_rest (l2))
				list_insert (right_options (g), list_pick (l2));
			l = list_start (right_options (g));
		}
		else
			l = list_rest (l);
	}
	return;
}

static void remove_dominated_options (list_type list, test_type test) {
	list_type list2, l, l2;
	game_type g, h;

	list2 = list_make();
	for (l = list_start (list); l != NIL; l = list_rest (l))
		if (! (dominates (list2, g = list_pick (l), test))) {
			for (l2 = list_start (list2); l2 != NIL;) {
				h = list_pick (l2);
				l2 = list_rest (l2);
				if ((* test) (g, h))
					list_delete (list2, h);
			}
			list_insert (list2, g);
		}
	list_clobber (list, list2);
}

static void evaluate (game_type g) {
	game_type l1, l2, l3, r1, r2, r3, h;
	int ls, rs;
	int maxlstar = -1, maxrstar = -1;
	Q_type n;
	int k;
	boolean is_num_equal = TRUE;
	boolean is_all_num_star_babyko = TRUE;
	boolean can_evaluate = TRUE;
	list_type l;
	
	ls = count_left_options (g);
	rs = count_right_options (g);
	l1 = ls >= 1 ? list_nth (left_options (g), 1) : FALSE;
	l2 = ls >= 2 ? list_nth (left_options (g), 2) : FALSE;
	l3 = ls >= 3 ? list_nth (left_options (g), 3) : FALSE;
	r1 = rs >= 1 ? list_nth (right_options (g), 1) : FALSE;
	r2 = rs >= 2 ? list_nth (right_options (g), 2) : FALSE;
	r3 = rs >= 3 ? list_nth (right_options (g), 3) : FALSE;
	n = l1 ? num_value (l1) : r1 ? num_value (r1) : Q_0;
	k = l1 ? babyko_value (l1) : r1 ? babyko_value (r1) : 0;
	for (l = list_start (left_options (g)); l != NIL; l = list_rest (l)) {
		h = list_pick (l);
		if (!Q_eq (num_value (h), n)) is_num_equal = FALSE;
		if (babyko_value (g) != k) can_evaluate = FALSE;
		if (!is_num_star_babyko (h)) is_all_num_star_babyko = FALSE;
		else if (star_value(h) > maxlstar) maxlstar = star_value (h);
		if (!is_evaluated (h)) can_evaluate = FALSE;
		if (has_ko (h)) set_flag (g, HAS_KO_FLAG);
	}
	for (l = list_start (right_options (g)); l != NIL; l = list_rest (l)) {
		h = list_pick (l);
		if (!Q_eq (num_value (h), n)) is_num_equal = FALSE;
		if (babyko_value (g) != k) can_evaluate = FALSE;
		if (!is_num_star_babyko (h)) is_all_num_star_babyko = FALSE;
		else if (star_value (h) > maxrstar) maxrstar = star_value (h);
		if (!is_evaluated (h)) can_evaluate = FALSE;
		if (has_ko (h)) set_flag (g, HAS_KO_FLAG);
	}
	if (!can_evaluate) return;
	else if (!l1 && !r1) nusb (Q_0,0,0,0);
	else if (!r1) nusb (Q_plus (n, Q_1),0,0,0);
	else if (!l1) nusb (Q_minus (n, Q_1),0,0,0);
	else if (is_num (l1) && is_num (r1) && !l2 && !r2) {
		if (Q_lt (num_value (l1), num_value (r1)))
			nusb (Q_divide (Q_plus (num_value (l1), num_value (r1)),Q_2),0,0,k);
		else if (is_num_equal) nusb (n,0,1,k);
	}
	else if (!is_num_equal) return;
	else if (l2 || r2) {
		if (is_num_babyko(r1) && !r2 && !l3
		    && (   (is_num_star_babyko(l1) && star_value(l1) == 1 && is_num_babyko(l2))
			|| (is_num_star_babyko(l2) && star_value(l2) == 1 && is_num_babyko(l1))))
			nusb (n,1,1,k);
		else if (is_num_babyko(l1) && !l2 && !r3
			 && (   (is_num_star_babyko(r1) && star_value(r1) == 1 && is_num_babyko(r2))
			     || (is_num_star_babyko (r2) && star_value(r2) == 1 && is_num_babyko(r1))))
			nusb (n,-1,1,k);
		else if (is_all_num_star_babyko && ls == rs
			 && maxlstar == (ls - 1) && maxrstar == (rs - 1))
			nusb (n,0,ls,k);
	}
	else if (is_num_babyko (l1) && !l2 && is_nusb (r1) && !r2 && (up_value (r1) >= 0))
		nusb (n, up_value(r1)+1, star_value(r1)^1,k);
	else if (is_num_babyko (r1) && !r2 && is_nusb (l1) && !l2 && (up_value (l1) <= 0))
		nusb (n, up_value(l1)-1, star_value(l1)^1,k);
	return;
}

static game_type simplify (game_type g) {
	list_type l;

	remove_dominated_options (left_options (g), ge);
	remove_dominated_options (right_options (g), le);
	for (l = list_start (left_options (g)); l != NIL; l = list_rest (l))
		if (has_ko (list_pick (l))) set_flag (g, HAS_KO_FLAG);
	for (l = list_start (right_options (g)); l != NIL; l = list_rest (l))
		if (has_ko (list_pick (l))) set_flag (g, HAS_KO_FLAG);
	reverse_reversals (g);
	clear_flag (g, HAS_KO_FLAG);
	remove_dominated_options (left_options (g), ge);
	remove_dominated_options (right_options (g), le);
	l = list_combine (left_options (g), right_options (g));
	if (hash_test (SIMPLIFY_KEY, l)) {
		game_free (g);
		g = hash_get_last();
		list_free (l);
	} else {
		hash_put (SIMPLIFY_KEY, l, g);
		evaluate (g);
		set_flag (g, CANONICAL_FLAG);
	}
	return g;
}

static int max_star (game_type g) {
	list_type l;
	int n = 0, s;
	if (is_num_up_star (g))
		if (up_value (g) == 0) return star_value (g);
		else return (1 | star_value (g));
 	for (l = list_start (left_options (g)); l != NIL; l = list_rest (l)) {
		s = max_star (list_pick (l));
		n = (n >= s ? n : s);
	}
	for (l = list_start (right_options (g)); l != NIL; l = list_rest (l)) {
		s = max_star (list_pick (l));
		n = (n >= s ? n : s);
	}
	return n;
}

game_type num_up_star_babyko (Q_type n, int u, int s, int k) {
	game_type g;
	list_type l, args;
	
	if (s < 0) s = -s;
	args = list_copy_3 (u,s,k);
	list_insert_unsorted (args, Q_p(n));
	list_insert_unsorted (args, Q_q(n));
	if (hash_test (NUM_UP_STAR_BABYKO_KEY, args)) {
		list_free (args);
		return hash_get_last();
	}
	if (u == 0 && s == 0) {		/* numbers */
		if (k == 0 || !Q_is_int (n))
			g = make_canonical ((!Q_is_int(n) || Q_gt(n,Q_0))		/* left */
					    ? list_copy_1 (nusb (Q (Q_p(n)-1, Q_q(n)),0,0,k))
					    : list_make()
					    ,
					    (!Q_is_int(n) || Q_lt (n, Q_0)) 		/* right */
					    ? list_copy_1 (nusb (Q (Q_p(n)+1, Q_q(n)),0,0,k))
					    : list_make());
		else if (Q_is_int (n)) g = int_babyko (Q_p(n), k);
	} else if (u == 0) {					/* num + star */
		g = nusb (n, 0, s-1, k);
		if (s > 1) l = list_copy (left_options (g));
		else l = list_make();
		list_insert (l, g);
		g = make_canonical (l, list_copy (l));
	} else if (u > 0) {
		if (u == 1 && s == 1)
			g = make_canonical (list_copy_2 (nusb (n,0,0,k), nusb (n,0,1,k)),
					    list_copy_1 (nusb (n,0,0,k)));
		else g = make_canonical (list_copy_1 (nusb (n,0,0,k)),
					 list_copy_1 (nusb (n,u-1,s^1,k)));
	} else /* if (u < 0) */ {
		if (u == -1 && s == 1)
			g = make_canonical (list_copy_1 (nusb (n,0,0,k)),
					    list_copy_2 (nusb (n,0,0,k), nusb (n,0,1,k)));
		else g = make_canonical (list_copy_1 (nusb (n,u+1,s^1,k)),
					 list_copy_1 (nusb (n,0,0,k)));
	}
	set_flag (g, NUSB_FLAG);
	game_cell[g]->Num = n;
	game_cell[g]->Up = u;
	game_cell[g]->Star = s;
	if (k != 0) {
		set_flag (g, HAS_KO_FLAG);
		set_flag (g, k == 1 ? BABYKO_FLAG : BABYKOBAR_FLAG);
	}
	hash_put (NUM_UP_STAR_BABYKO_KEY, args, g);
	return g;
}

game_type int_ko_star (int n, int k, int s) {
	game_type g, h;
	int i;

	if (s < 0) return int_ko_star (n, k, -s);
	else if (k > 1) return int_ko_star (n+1, k-3, s);
	else if (k < -1) return int_ko_star (n-1, k+3, s);
	else if (k == 0) return num_up_star (Q_int (n), 0, s);
	if (hash_test (INT_KO_STAR_KEY, list_3 (n, k, s)))
		return hash_get_last();
	i = (k == 1 ? n : n-1);
	g = new_game();
	h = new_game();

	game_cell[g]->left = list_copy_1 (h);
	game_cell[g]->right = list_copy_1 (plus (g_int (i), star(s)));
	game_cell[g]->Num = Q_int (i);
	game_cell[g]->Star = s;
	set_flag (g, CANONICAL_FLAG);
	set_flag (g, KO_FLAG);
	set_flag (g, HAS_KO_FLAG);

	game_cell[h]->left = list_copy_1 (plus (g_int (i+1), star(s)));
	game_cell[h]->right = list_copy_1 (g);
	game_cell[h]->Num = Q_int (i+1);
	game_cell[h]->Star = s;
	set_flag (h, CANONICAL_FLAG);
	set_flag (h, KOBAR_FLAG);
	set_flag (h, HAS_KO_FLAG);

	hash_put (INT_KO_STAR_KEY, list_copy_3 (i, 1, s), g);
	hash_put (INT_KO_STAR_KEY, list_copy_3 (i+1, -1, s), h);
	if (k == 1) return g;
	else return h;
}

game_type int_babyko (int n, int k) {
	game_type g, h;
	if (k > 1) return int_babyko (n+1, k-3);
	else if (k < -1) return int_babyko (n-1, k+3);
	else if (k == 0) return g_int (n);
	if (hash_test (INT_BABYKO_KEY, list_2 (n, k)))
		return hash_get_last();
	g = new_game();
	h = new_game();

	game_cell[g]->left = list_copy_1 (h);
	game_cell[g]->right = list_copy_1 (g_int (n+1));
	game_cell[g]->Num = Q_int (n);
	set_flag (g, CANONICAL_FLAG);
	set_flag (g, NUSB_FLAG);
	set_flag (g, BABYKO_FLAG);
	set_flag (g, HAS_KO_FLAG);

	game_cell[h]->left = list_copy_1 (g_int (n-1));
	game_cell[h]->right = list_copy_1 (g);
	game_cell[h]->Num = Q_int (n);
	set_flag (h, CANONICAL_FLAG);
	set_flag (h, NUSB_FLAG);
	set_flag (h, BABYKOBAR_FLAG);
	set_flag (h, HAS_KO_FLAG);

	hash_put (INT_BABYKO_KEY, list_copy_2 (n,  1), g);
	hash_put (INT_BABYKO_KEY, list_copy_2 (n, -1), h);
	if (k == 1) return g;
	else return h;
}
	
void init (void) {
	list_init();
	hash_init();
	zero = make(list_make(), list_make());
	one = g_int (1);
	two = g_int (2);
	neg1 = g_int (-1);
	one_star = plus (g_int (1), star (1));
	ko = int_ko_star (0,1,0);
	kobar = int_ko_star (0,-1,0);
	babyko = int_babyko (0,1);
	babykobar = int_babyko (0,-1);
	mult (zero, one_star);
	heat (zero, one);
	overheat (zero, one_star, one);
	strcpy (chill_string, "g[1]\n");
	strcpy (warm_string, "g.1*\n");
	dummy_var = list_0;
	period = 2;
	saltus = zero;
}

void game_clear (void)
{
	game_type g;

	for (g = game_cells; g >= 1; g--)
		game_free (g);
	list_free (unit_neg_adder);
	list_free (unit_adder);
	unit_adder = unit_neg_adder = NIL;
}

void reinit (void) {
	hash_clear(UNEXPENDABLE_KEYS | EXPENDABLE_KEYS | OTHER_KEYS, 100);
	corridor_clear();
	game_clear();
	zero = make(list_make(), list_make());
	one = g_int (1);
	two = g_int (2);
	neg1 = g_int (-1);
	one_star = plus (g_int (1), star (1));
	ko = int_ko_star (0,1,0);
	kobar = int_ko_star (0,-1,0);
	babyko = int_babyko (0,1);
	babykobar = int_babyko (0,-1);
	mult (zero, one_star);
	heat (zero, one);
	overheat (zero, one_star, one);
}

void clean (void) {
	hash_clear (EXPENDABLE_KEYS | OTHER_KEYS, 100);
	corridor_clear();
}	

game_type make (list_type left, list_type right) {
	game_type g;

	g = new_game();
	game_cell[g]->left = left;
	game_cell[g]->right = right;
	g = simplify(g);
	return g;
}

game_type make_canonical (list_type left, list_type right) {
	game_type g;
	list_type l;

	l = list_combine (left, right);
	if (hash_test (SIMPLIFY_KEY, l)) {
		g = hash_get_last();
		list_free (l);
		list_free (left);
		list_free (right);
	} else {
		g = new_game();
		game_cell[g]->left = left;
		game_cell[g]->right = right;
		set_flag (g, CANONICAL_FLAG);
		hash_put (SIMPLIFY_KEY, l, g);
	}
	return g;
}

boolean is_var (game_type g) {
	var_type v;

	v = game_cell[g]->var;
	if (v == NIL) return FALSE;
	if (eq (get_var (v), g)) return TRUE;
	else {
		list_free (v);
		game_cell[g]->var = NIL;
		return FALSE;
	}
}

boolean is_tiny (game_type g) {
	game_type h;
	return (count_left_options (g) == 1
		&& count_right_options (g) == 1
		&& eq (nth_left_option (g, 1), zero)
		&& (h = nth_right_option (g, 1), TRUE)
		&& count_left_options (h) == 1
		&& count_right_options (h) == 1
		&& eq (nth_left_option (h, 1), zero)
		&& lt (nth_right_option (h, 1), zero));
}
boolean is_miny (game_type g) {return is_tiny (negative (g));}

static int on_exponent_ (game_type g) {
	game_type h;
	int i;

	if (count_left_options (g) != 1 || count_right_options (g) != 1) return 0;
	h = nth_right_option (g, 1);
	g = nth_left_option (g, 1);
	for (i = 1;; i++) {
		if (eq (g, zero)) return i;
		if (count_left_options (g) != 1 || count_right_options (g) != 1) return 0;
		if (!eq (h, nth_right_option (g, 1))) return 0;
		g = nth_left_option (g, 1);
	}
}

boolean is_on (game_type g) {return (on_exponent (g) > 1) && !is_nusb(g);}
int on_exponent (game_type g) {
	int i;

	if ((i = on_exponent_ (g)) > 0) return i;
	else return on_exponent_ (negative (g));
}

game_type on_value (game_type g) {
	if (on_exponent_ (g) > 1) return vs (zero, nth_right_option (g, 1));
	else if (on_exponent_ (g) < 1) return negative (on_value (negative (g)));
	else return g;
}

boolean is_off (g) game_type g; {return is_on (negative (g));}

game_type tiny_value(game_type g) {g=minus (g, num (mean (g)));
  return negative (nth_right_option(nth_right_option (g,1),1));}
game_type miny_value(game_type g) {g=minus (g, num (mean(g)));
  return nth_left_option (nth_left_option (g,1),1);}
var_type var_value (game_type g) {return is_var(g) ? game_cell[g]->var : NO_GAME;}
int ko_value (game_type g) {return _flags(g)&KO_FLAG ? 1 : _flags(g)&KOBAR_FLAG ? -1 : 0;}
int babyko_value (game_type g) {return _flags(g)&BABYKO_FLAG ? 1 : _flags(g)&BABYKOBAR_FLAG ? -1 : 0;}

boolean is_canonical(game_type g) {return _flags(g) & CANONICAL_FLAG;}
boolean is_evaluated(game_type g) {return _flags(g) & EVALUATED_FLAG;}
boolean is_int(game_type g) {return is_num (g) && Q_is_int (num_value (g));}
boolean is_num(game_type g) {return is_num_star (g) && star_value (g) == 0;}
boolean is_up(game_type g) {return is_up_star (g) && star_value (g) == 0;}
boolean is_star(game_type g) {return is_up_star (g) && up_value (g) == 0;}
boolean is_num_up_star(game_type g) {return is_nusb(g)&&!babyko_value(g);}
boolean is_num_star(game_type g) {return is_num_up_star (g) && up_value (g) == 0;}
boolean is_num_up(game_type g) {return is_num_up_star (g) && star_value (g) == 0;}
boolean is_up_star(game_type g) {return is_num_up_star(g) && Q_eq (num_value (g), Q_0);}
boolean is_int_ko_star(game_type g) {return (is_num_star(g)&&Q_is_int(num_value(g))&&star_value(g)<=1) || ko_value(g);}
boolean is_int_ko(game_type g) {return (is_int_ko_star(g) && star_value (g) == 0);}
boolean is_num_up_star_babyko (game_type g) {return _flags (g) & NUSB_FLAG;}
boolean is_num_star_babyko (game_type g) {return is_nusb (g) && up_value (g) == 0;}
boolean is_num_babyko (game_type g) {return is_num_star_babyko (g) && star_value (g) == 0;}
boolean is_int_babyko(game_type g) {return is_num_babyko (g) && Q_is_int (num_value (g));}
boolean has_ko(game_type g) {return _flags (g) & HAS_KO_FLAG;}
	
boolean is_all_small(game_type g) {
	int i,n;

	if (is_evaluated (g))
		return (eq (num_value (g), Q_0) && !has_ko (g));
	else {
		for (i=1,n=count_left_options(g); i<=n; i++)
			if (!is_all_small (nth_left_option (g,i))) return FALSE;
		for (i=1,n=count_right_options(g); i<=n; i++)
			if (!is_all_small (nth_right_option (g,i))) return FALSE;
	}
	return TRUE;
}

char *var_sprintn (char *s, var_type v, int n) {
	list_type l;
	char *t;

	t = s;
	for (l = list_start (v); l != NIL && n > 0; l = list_rest(l), s++, n--)
		*s = list_pick (l) & 255;
	*s = '\0';
	return t;
}

game_type plus (game_type g, game_type h) {
	list_type l1, l2, left, right;
	game_type result;

	if (is_nusb (g) && is_nusb (h))
		return nusb (Q_plus (num_value (g), num_value (h)),
			     up_value (g) + up_value (h),
			     star_value (g) ^ star_value (h),
			     babyko_value (g) + babyko_value (h));
	else if (eq (g, zero)) return h;
	else if (eq (h, zero)) return g;
	else if (is_int_ko_star (g) && is_int_ko_star (h))		/*KLUGED KO*/
		return int_ko_star (int_value (g) + int_value (h),
				    ko_value (g) + ko_value (h),
				    star_value (g) ^ star_value (h));
	if (hash_test (PLUS_KEY, list_unordered_2(g,h)))
		return hash_get_last ();
	if (is_num (g) || (is_int_ko_star (g) && has_ko (g)))		/*KLUGED KO*/
		result = make (list_mapcl (plus, g, left_options (h)),
			       list_mapcl (plus, g, right_options (h)));
	else if (is_num (h) || (is_int_ko_star (h) && has_ko (h)))	/*KLUGED KO*/
		result = make (list_maplc (plus, left_options (g), h),
			       list_maplc (plus, right_options (g), h));
	else {
		l1 = list_maplc (plus, left_options (g), h);
		l2 = list_mapcl (plus, g, left_options (h));
		left = list_union (l1, l2);
		list_free (l1);
		list_free (l2);
		l1 = list_maplc (plus, right_options (g), h);
		l2 = list_mapcl (plus, g, right_options (h));
		right = list_union (l1, l2);
		list_free (l1);
		list_free (l2);
		result = make (left, right);
	}
	hash_put (PLUS_KEY, list_copy_unordered_2(g,h), result);
	return result;
}


game_type negative (game_type g) {
	game_type result;

	if (is_nusb (g))				/*KLUGED KO*/
		return nusb (Q_negative (num_value (g)), - up_value (g),
			     star_value (g), - babyko_value (g));
	else if (is_int_ko_star (g))			/*KLUGED KO*/
		return int_ko_star (- int_value (g), - ko_value (g), star_value (g));
	if (hash_test (NEGATIVE_KEY, list_1(g)))
		result = hash_get_last();
	else {
		result = make (list_map (negative, right_options (g)),
			       list_map (negative, left_options (g)));
		hash_put (NEGATIVE_KEY, list_copy_1(g), result);
	}
	return result;
}

game_type minus (game_type g, game_type h) {
	return plus (g, negative (h));
}

game_type times (int n, game_type g) {
	if (n == 0) return zero;
	else if (n < 0) return times (-n, negative (g));
	else return plus (g, times (n-1, g));
}

boolean ge (game_type g, game_type h) {
	list_type l, r;
	boolean result = TRUE;
	int t;
	game_type d, gr, hl;

	if (eq (g, h)) return TRUE;
	else if (is_nusb (g) && is_nusb (h)) {
		d = minus (g, h);
		if (babyko_value (d) == -1)
			d = minus (d, num (Q (1,2)));	/* 0 <= k <= 1/2 */
		if      (Q_gt (num_value (d), Q_0)) return TRUE;
		else if (Q_lt (num_value (d), Q_0)) return FALSE;
		else if ((t = up_value (d)) < 0) return FALSE;
		else if (t >= 2) return TRUE;
		else if (t == 1) return star_value (d) != 1;
		else return star_value (d) == 0;
	} else if (is_num (h) && !has_ko (g)) {
		for (l = list_start (right_options (g)); l != NIL; l = list_rest(l))
			if (ge (h, list_pick (l))) return FALSE;
		return TRUE;
	} else if (is_num (g) && !has_ko (h)) {
		for (l = list_start (left_options (h)); l != NIL; l = list_rest(l))
			if (ge (list_pick (l), g))
				return FALSE;
		return TRUE;
	} else if (is_int_ko_star (g) && is_int_ko_star (h)) {
		d = minus (g, h);
		return (int_value (d) > 0 || eq (d, zero) || eq (d, int_ko_star (0,1,1)));
	} else if (hash_test (GE_KEY, list_2(g,h)))
		return  (boolean) hash_get_last();
	if (has_ko (g) || has_ko (h)) {		/*KLUGED KO*/
		for (r = list_start (right_options (g));
		     r != NIL; r = list_rest (r)) {
			t = FALSE;
			gr = list_pick (r);
			for (l=list_start(extended_left_options(gr)); l!=NIL; l=list_rest(l))
				if (ge (list_pick (l), h)) {t = TRUE; break;}
			if (!t)	for (l=list_start(extended_right_options(h)); l!=NIL; l=list_rest(l))
				if (ge (gr, list_pick (l))) {t = TRUE; break;}
			if (!t) {result = FALSE; break;}
		}
		if (result) for (r = list_start (left_options (h));
				 r != NIL; r = list_rest (r)) {
			t = FALSE;
			hl = list_pick (r);
			for (l=list_start(extended_right_options(hl)); l!=NIL; l=list_rest(l))
				if (ge (g, list_pick (l))) {t = TRUE; break;}
			if (!t) for (l=list_start(extended_left_options(g)); l!=NIL; l=list_rest(l))
				if (ge (list_pick (l), hl)) {t = TRUE; break;}
			if (!t) {result = FALSE; break;}
		}
	} else {
		for (l = list_start (left_options (h)); l != NIL; l = list_rest(l))
			if (ge (list_pick (l), g)) {result = FALSE; break;}
		if (result) for (l = list_start (right_options (g)); l != NIL; l = list_rest(l))
			if (ge (h, list_pick (l))) {result = FALSE; break;}
	}
	if (is_canonical (g) && is_canonical (h))
		hash_put (GE_KEY, list_copy_2(g,h), (int) result);
	return result;
}
boolean le(game_type g, game_type h) {return (ge (h, g));}
boolean eq(game_type g, game_type h) {return g == h;}
boolean gt(game_type g, game_type h) {return g != h && ge(g,h);}
boolean lt(game_type g, game_type h) {return g != h && ge(h,g);}


char *compare (game_type g, game_type h) {
	boolean a, b;
	a = ge (g, h);
	b = le (g, h);
	if (a && b) return "=";
	else if (a && !b) return ">";
	else if (!a && b) return "<";
	else return "<>";
}

boolean lmove_ok (game_type g, game_type h) {
	list_type l;
	for (l = list_start (left_options (g)); l != NIL; l = list_rest (l))
		if (ge (h, list_pick (l))) return TRUE;
	return FALSE;
}

boolean rmove_ok (game_type g, game_type h) {
	list_type l;
	for (l = list_start (right_options (g)); l != NIL; l = list_rest (l))
		if (le (h, list_pick (l))) return TRUE;
	return FALSE;
}


Q_type left_stop (game_type g) {
	Q_type n, m;
	list_type l;

	if (is_evaluated (g)) return num_value (g);
	l = list_start (left_options (g));
	if (l == NIL) return Q_minus (right_stop (g), Q_1);	/*KLUGED KO*/
	else n = right_stop (list_pick (l));
	while ((l = list_rest (l)) != NIL) {
		m = right_stop (list_pick (l));
		n = (Q_gt (m, n) ? m : n);
	}
	return n;
}

Q_type right_stop (game_type g) {
	Q_type n, m;
	list_type l;

	if (is_evaluated (g)) return num_value (g);
	l = list_start (right_options (g));
	if (l == NIL) return Q_plus (left_stop (g), Q_1);	/*KLUGED KO*/
	n = left_stop (list_pick (l));
	while ((l = list_rest (l)) != NIL) {
		m = left_stop (list_pick (l));
		n = (Q_lt (m, n) ? m : n);
	}
	return n;
}

list_type incentives (game_type g) {
	list_type l1, l2, l3;

	l1 = left_incentives (g);
	l2 = right_incentives (g);
	l3 = list_union (l1, l2);
	remove_dominated_options (l3, ge);
	list_free (l1);
	list_free (l2);
	return l3;
}
	
list_type left_incentives (game_type g) {return is_int(g) ? list_copy_1 (neg1) : list_maplc (minus, left_options (g), g);}
list_type right_incentives (game_type g) {return is_int(g) ? list_copy_1 (neg1) : list_mapcl (minus, g, right_options (g));}

game_type cool (game_type g, Q_type n) {
	Q_type low, high, try;
	list_type left, right;
	list_type l;
	game_type gtry, h;
	boolean cold=FALSE;
	
	if (is_num (g) || Q_eq (n, Q_0)) {
		frozen_temperature = Q_0;
		return g;
	} else if (is_evaluated (g)) {
		frozen_temperature = Q_0;
		h = num (num_value (g));
		if (ko_value (g) == 0) return h;
		else return plus (h, ko_value (g) == 1 ? babyko : babykobar);
	}
	/* The following code is complicated by the fact that we want */
	/* to guarantee that the hash table entries are expendable: */
	/* One of FREEZE_KEY and TEMPERATURE_KEY may have been axed! */
	else if (hash_test (FREEZE_KEY, list_1 (g))) {
		h = hash_get_last();
		if (hash_test (TEMPERATURE_KEY, list_1(g)))
			if (Q_ge (n, frozen_temperature = num_value (hash_get_last())))
				if (Q_eq (n, frozen_temperature)) return h;
				else return num (left_stop (h));
		if (Q_eq (n, Q_infinity)) return num (left_stop (h));
		hash_delete (FREEZE_KEY, list_1 (g));
	}
	if (hash_test (TEMPERATURE_KEY, list_1(g))) hash_delete (TEMPERATURE_KEY, list_1(g));
#if Q_SIZE == INT_SIZE
	if (hash_test (COOL_KEY, list_2 (g, n)))
		return hash_get_last();
#else
        if (hash_test (COOL_KEY, list_3 (g, Q_p(n), Q_q(n))))
                return hash_get_last();
#endif
	if (has_ko (g) && Q_le (left_stop (g), right_stop (g))) cold = TRUE;	/*KLUGED KO*/
	low = Q_neg1;
	high = n;
	try = (Q_eq (high, Q_infinity) ? Q_0 : high);
	for (;;) {
		left = list_make();
		right = list_make();
		gtry = num (try);
		
		for (l = list_start (left_options (g)); l != NIL; l = list_rest (l))
			list_insert (left, minus (cool (list_pick (l), try), gtry));
		for (l = list_start (right_options (g)); l != NIL; l = list_rest (l))
			list_insert (right, plus (cool (list_pick (l), try), gtry));
		h = make (left, right);
		if (!is_num (h) && (   Q_le (left_stop (h), right_stop (h)) /* frozen */
				    || Q_eq (try, n)			 /* cooled enough */
				    || cold)) break;			/*KLUGED KO*/
		if (is_num (h)) high = try;
		else low = try;
		try = Q_simplest_number (low, high);
	}
	frozen_temperature = cold ? Q_0 : try;				/*KLUGED KO*/
	if (Q_eq (left_stop (h), right_stop (h))) {
		hash_put (FREEZE_KEY, list_copy_1 (g), h);
		hash_put (TEMPERATURE_KEY, list_copy_1 (g), num (frozen_temperature));
	} else
#if Q_SIZE == INT_SIZE
		hash_put (COOL_KEY, list_copy_2 (g, n), h);
#else
		hash_put (COOL_KEY, list_copy_3 (g, Q_p(n), Q_q(n)), h);
#endif
	if (Q_gt (n, frozen_temperature)) return num (left_stop (h));
	else return h;
}
Q_type temperature (game_type g) {cool (g, Q_infinity); return frozen_temperature;}
game_type freeze (game_type g) {return cool (g, temperature (g));}
Q_type mean (game_type g) {return left_stop (freeze (g));}

game_type overheat (game_type g, game_type s, game_type t) {
	list_type left;
	list_type right;
	list_type l;
	game_type h;
	static game_type last_s, last_t;

	if (s == NO_GAME) s = last_s; else last_s = s;
	if (t == NO_GAME) t = last_t; else last_t = t;
	if (is_int (g)) return times (int_value (g), s);
	else if (is_int_babyko (g)) return plus (star(1), plus (times (int_value (g), s),	/*KLUGED KO*/
								babyko_value (g) == 1 ? ko : kobar));
	else if (hash_test (HEAT_KEY, list_3 (g, s, t)))
		return hash_get_last();
	left = list_make();
	right = list_make();
	for (l = list_start (left_options (g)); l != NIL; l = list_rest (l))
		list_insert (left, plus (overheat (list_pick (l), s, t), t));
	for (l = list_start (right_options (g)); l != NIL; l = list_rest (l))
		list_insert (right, minus (overheat (list_pick (l), s, t), t));
	h = make (left, right);
	hash_put (HEAT_KEY, list_copy_3 (g, s, t), h);
	return h;
}

game_type heat (game_type g, game_type t) {
	list_type left, right, l;
	game_type h;
	static game_type last_t;

	if (t == NO_GAME) t = last_t; else last_t = t;
	if (is_num (g)) return g;
	else if (hash_test (HEAT_KEY, list_2 (g, t)))
		return hash_get_last ();
	left = list_make();
	right = list_make();
	for (l = list_start (left_options (g)); l != NIL; l = list_rest (l))
		list_insert (left, plus (heat (list_pick (l), t), t));
	for (l = list_start (right_options (g)); l != NIL; l = list_rest (l))
		list_insert (right, minus (heat (list_pick (l), t), t));
	h = make(left, right);
	hash_put (HEAT_KEY, list_copy_2 (g, t), h);
	return h;
}

game_type _mult (game_type g) {return mult (g, NO_GAME);}
game_type mult (game_type g, game_type u) {
	game_type h;
	list_type l, left, right;

	if (u != NO_GAME) {
		if (!gt (u, zero)) myerror ("Unit for norton multiply not greater than 0");
		unit = u;
		l = incentives (unit);
		if (unit_adder != NIL) {
			list_free (unit_adder);
			list_free (unit_neg_adder);
		}
		unit_adder = list_mapcl (plus, unit, l);
		unit_neg_adder = list_map (negative, unit_adder);
		list_free (l);
	}
	if (is_int (g)) return times (int_value (g), unit);
	else if (is_int_babyko (g)) return plus (star (1), plus (times (int_value (g), unit),	/*KLUGED KO*/
								 babyko_value (g) == 1 ? ko : kobar));
	else if (hash_test (MULT_KEY, list_2 (g, unit)))
		return hash_get_last();
	left = list_map (_mult, left_options (g));
	right = list_map (_mult, right_options (g));
	h = make (list_mapcomb (plus, left, unit_adder),
		  list_mapcomb (plus, right, unit_neg_adder));
	list_free (left);
	list_free (right);
	hash_put (MULT_KEY, list_copy_2 (g, unit), h);
	return h;
}

static game_type aw (game_type g) {
	list_type left, right, l;
	game_type g0, h;
	int try;

	if (hash_test (AW_KEY, list_1 (g)))
		return hash_get_last();
	left = list_make();
	right = list_make();

	for (l = list_start (left_options (g)); l != NIL; l = list_rest (l))
		list_insert (left, minus (aw (list_pick (l)), two));
	for (l = list_start (right_options (g)); l != NIL; l = list_rest (l))
		list_insert (right, plus (aw (list_pick (l)), two));
	g0 = make (left, right);
	if (!is_int (g0)) {
		hash_put (AW_KEY, list_copy_1 (g), g0);
		return g0;
	} else if (gt (g, _rstar_)) {
		for (h = neg1, try = -1;;h = g_int (++try))
			if ((l = list_start (right_options (g))) != NIL) {
				for (; l != NIL; l = list_rest (l))
					if (ge (h, aw (list_pick (l)))) {
						hash_put (AW_KEY, list_copy_1 (g), h = g_int (try+1));
						return h;
					}
			} else {
				hash_put (AW_KEY, list_copy_1 (g), h = g_int (try+1));
				return h;
			}
	} else if (lt (g, _rstar_)) {
		for (h = one, try = 1;; h = g_int (--try))
			if ((l = list_start (left_options (g))) != NIL) {
				for (; l != NIL; l = list_rest (l))
					if (le (h, aw (list_pick (l)))) {
						hash_put (AW_KEY, list_copy_1 (g), h = g_int (try-1));
						return h;
					}
			} else {
				hash_put (AW_KEY, list_copy_1 (g), h = g_int (try - 1));
				return h;
			}
	} else {
		hash_put (AW_KEY, list_copy_1 (g), g0);
		return g0;
	}
}

game_type kite (int n) {
	int m;
	/* returns a green hackenbush string of length |n| with a kite */
	/* whose color is blue if n>0, red if n<0 and none if n=0 */
	if (n == 0) m = 0;
	else if (n > 0) m = n++;
	else m = (n=-n) + 1;
	return make (list_copy (left_options (star (n))),
		     list_copy (right_options (star (m))));
}

game_type atomic_weight (game_type g) {
	game_type result, h, diff_kites;

	if (! (Q_eq (left_stop (g), Q_0) && Q_eq (right_stop (g), Q_0))) {
		myerror ("atomic weight: Games is not infinitesimal.");
		return zero;
	}
	_rstar_ = star (max_star (g) + 1);
	result = aw (g);
	if (!is_all_small (g)) { /* Test if the result holds despite */
				 /* no support by the theory */
		h = minus (g, mult (result, up(1)));
		diff_kites = minus (kite(1), kite(2));
		if (le (h, diff_kites) && ge (h, negative (diff_kites)))
/*			myerror ("atomic weight: Game not all small, but result is ok!")*/;
		else
			myerror ("atomic weight: Game not all small.  Do not trust result.");
	}
	return result;
}

list_type make_var (char *s) {
	int k;
	var_type v;
	
	v = list_make();
	for (k=0; *s!='\0'; s++, k++)
		list_insert (v, (k<<8) | *s);
	return v;
}

void set_var (var_type v, game_type g) {
	hash_delete (VAR_KEY, v);
	if (!list_equal (v, dummy_var)) {
		list_free (game_cell[g]->var);
		game_cell[g]->var = list_copy (v);
	}
	hash_put (VAR_KEY, v, g);
}

game_type get_var (var_type v) {
	if (hash_test (VAR_KEY, v)) return hash_get_last();
	else {
		myerror ("get_var: Variable not set.");
		return NO_GAME;
	}
}

unsigned long game_stat (void) {
	unsigned long n=0, t=0, v=0, nv=0;
	game_type g;
	game_cell_type l;

	for (g = 1; g <= game_cells; g++) {
		n = (n
		     + (left_options (g) ? count_extended_left_options (g) + 1 : 0)
		     + (right_options (g) ? count_extended_right_options (g) + 1 : 0));
		if (game_cell[g]->var != NIL) {
			nv++;
			v = v + list_length (game_cell[g]->var) + 1;
		}
	}
	t = list_length (unit_adder) + list_length (unit_neg_adder) + 2;
	printf ("game option lists: %lu; mult lists: %lu; variable lists: %lu\n", n, t, v);
	printf ("variables: %lu; ", nv);
	for (nv=0, l = game_free_list; l != GNIL; l = l->next) nv++;
	printf ("games: %lu; total game cells: %lu\n", game_cells, game_cells + nv);
	return n + t + v;
}

game_type zeros_vs (game_type g, int n)
{
	if (n == 0) return g;
	else if (n < 0) return negative (zeros_vs (g, -n));
	else return vs (zero, zeros_vs (g, n-1));
}
game_type vs_zeros(game_type g, int n) {return zeros_vs (negative (g), -(n));}
game_type vs(game_type g, game_type h) {return make (list_copy_1 (g), list_copy_1 (h));}
list_type left_options (game_type g) {return ko_value (g) || babyko_value (g) ? list_0 : game_cell[g]->left;}
list_type right_options(game_type g) {return ko_value (g) || babyko_value (g) ? list_0 : game_cell[g]->right;}

game_type on (game_type g, int n) {
	game_type h;
	
	if (! (count_left_options(g)==1 && count_right_options(g)==1)) {
		myerror ("on: Game not in correct form for on");
		return zero;
	}
	else if (n == 0) return zero;
	else if (n < 0) return negative (on (g, -n));
	else if (eq (nth_left_option (g, 1), zero)) {
		h = vs (on (g, n-1), nth_right_option (g, 1));
		if (!eq (h, plus (on (g, n-1), gexp (g, n))))
			myerror ("on: on and gexp didn't match");
		return h;
	} else if (eq (nth_right_option (g, 1), zero))
		return negative (on (negative (g), n));
	else {
		myerror ("on: Game not in correct form for on");
		return zero;
	}
}
game_type off (game_type g, int n) {return on (g, -n);}
game_type gexp (game_type g, int n) {
	game_type h;
	int i;

	if (! (count_left_options(g)==1 && count_right_options(g)==1)) {
		myerror ("gexp: Game not in correct form for gexp");
		return zero;
	}
	else if (n == 0) return zero;
	else if (n < 0) return negative (gexp (g, -n));
	else if (eq (nth_left_option (g, 1), zero)) {
		for (h = nth_right_option (g, 1), i=1; i<n; i++)
			h = minus (h, gexp (g, i));
		return vs (zero, h);
	} else if (eq (nth_right_option (g, 1), zero)) return negative (gexp (negative (g), n));
	else {
		myerror ("gexp: Game not in correct form for gexp");
		return zero;
	}
}

game_type tweak(game_type g)
{
    return heat(cool(g, (Q_type) 1), num((Q_type) 1));
}
